/* FILE: mesh.h                 -*-Mode: c++-*-
 *
 * Abstract mesh class and children
 *
 */

#ifndef _VF_MESH
#define _VF_MESH

#include "nb.h"

#include "ptsearch.h"

/* End includes */     /* Optional directive to build.tcl */

//////////////////////////////////////////////////////////////////////////
// Extension of LocatedVector class used for display purposes.
// Perhaps both this and LocatedVector should we written up via
// templates?
struct Vf_DisplayVector
{
public:
  Nb_Vec3<REAL4> location;
  Nb_Vec3<REAL4> value;
  REAL4m shade;
  Vf_DisplayVector() {
    location.Set(REAL4(0.),REAL4(0.),REAL4(0.));
    value.Set(REAL4(0.),REAL4(0.),REAL4(0.));
    shade=REAL4(0.);
  }
  Vf_DisplayVector(const Nb_Vec3<REAL4>& _location,
		const Nb_Vec3<REAL4>& _value,REAL4m _shade) {
    location=_location; value=_value; shade=_shade;
  }
};


//////////////////////////////////////////////////////////////////////////
// Vf_Mesh (abstract) class

class Vf_Mesh_Index
{ // Index class used by Vf_Mesh::GetFirst() and ::GetNext()
  // Children of Vf_Mesh should define a compatible index class
  // as a child of this class.
public:
  virtual ~Vf_Mesh_Index() {}
};

class Vf_Mesh
{
private:
  const char * const MeshTypename;
  char *Filename;
  char *Title;
  char *Description;
  char *MeshUnit;  // Measurement unit for grid; e.g., "m"
  char *ValueUnit;

  // The following 3 functions are left undefined on purpose
  Vf_Mesh();                   // Default constructor
  Vf_Mesh(const Vf_Mesh&);     // Copy constructor
  Vf_Mesh& operator=(Vf_Mesh &rhs); // Assignment op
protected:
  void SetFilename(const char* _Filename);
  void SetDescription(const char* _Description);
  void SetMeshUnit(const char* _MeshUnit);
  void SetValueUnit(const char* _ValueUnit);
  BOOL UseMagHints;
  REAL8 MinMagHint,MaxMagHint;  // Hint values for min &
  /// max magnitude of all vectors in mesh.  If UseMagHints
  /// is true, then any vector smaller than MinMagHint is
  /// set to 0 length, and any larger than MaxMagHint is
  /// rescaled to have magnitude equal to MaxMagHint.
  REAL8 ValueMultiplier;
public:
  const char* GetName() const;
  const char* GetTitle() const;
  const char* GetDescription() const;
  void SetTitle(const char* _Title);
  // NOTE: "Filename", "Title" & "Description" are owned and managed
  // by class Vf_Mesh, and the returned address will become invalid if
  // SetFilename/Title/Descripion is called or if the mesh is
  // destroyed.  Calling routines will want to copy the returned
  // string into local memory if it is needed for an extended time.

  const char* GetMeshUnit() const;
  const char* GetValueUnit() const;
  REAL8 GetValueMultiplier() const { return ValueMultiplier; }

  BOOL SetMagHints(REAL8 minmag,REAL8 maxmag) {
    if(minmag<=0. || minmag>maxmag) return 0; // Error: Bad input values
    UseMagHints=1; MinMagHint=minmag; MaxMagHint=maxmag;
    return 1;
  }
  void GetMagHints(BOOL& active,REAL8& minmag,REAL8& maxmag) const {
    active=UseMagHints; minmag=MinMagHint; maxmag=MaxMagHint;
  }
  void UnsetMagHints() { UseMagHints=0; }

  const char* GetMeshType() const { return MeshTypename; }

  INT4m GetRange(Nb_BoundingBox<REAL4> &range) const;
  virtual INT4m GetPreciseRange(Nb_BoundingBox<REAL8> &range) const =0;
  /// Data + boundary range.  Tightness not required, but should
  /// be close.

  INT4m GetDataRange(Nb_BoundingBox<REAL4> &range) const;
  virtual INT4m GetPreciseDataRange(Nb_BoundingBox<REAL8> &range) const =0;
  /// Data range only, no boundary.  Tightness not required, but
  /// should be close.

  virtual INT4m GetBoundaryList(Nb_List< Nb_Vec3<REAL4> > &boundary_list)
    const =0;
  /// Used to draw outline polygon in display.  Should probably enclose
  /// all data points.

  virtual REAL4m GetApproximateCellWidth() const =0;
  virtual Nb_Vec3<REAL4>  GetApproximateCellDimensions() const =0;
  /// This should be an "average" distance between nearest
  /// neighbor points.  Used to size arrows in display.

  virtual REAL4m GetPreferredStepSize(REAL4m step) const =0;
  /// Import "step" and return value is in mesh units.  For regular grids
  /// the return value should probably be an integral multiple of
  /// "GetApproximateCellWidth()".

  virtual REAL4m GetSubsampleGrit(void) const =0;
  /// Used to set increment on subsample scale widget in display control
  /// window.  Units are RELATIVE to "GetApproximateCellWidth()".  For
  /// regular grids it doesn't make sense to sample at a rate smaller
  /// than the grid lattice spacing, so this value should be 1 (provided
  /// GetApproximateCellWidth() returns the lattice spacing constant),
  /// but for irregular meshes, a smaller value may make perfect sense.
  ///  GetSubsampleGrit() should return a value >0 and <=1.

  virtual INT4m ColorQuantityTypes(Nb_List<Nb_DString> &types) const {
    types.Clear();
    return 0;
  }
  /// List of strings detailing color quantities supported.  Return
  /// value is the number of elements in the list.  Non-trivial meshes
  /// should support at least "x", "y", "z", and "none".

  virtual REAL4m GetVecShade(const char* /* colorquantity */,
                             const Nb_Vec3<REAL4>& /* v */) const {
    return REAL4m(-1.);
  }
  /// Given a vector and a plotting quantity, returns the "shade" that vector
  /// would be colored.  The return value is between 0. and 1. is the plotting
  /// quantity is computable for a single vector, or -1. if colorquantity is
  /// unrecognized or not computable.  This routine is intended to be used for
  /// coloring coordinate axes, and makes sense if colorquantity is "x", "y"
  /// or "z".  For a quantity like "div" there is no logical coloring of a
  /// single vector, and so -1. will be returned.  "none" should also return
  /// -1.
  ///  NB: Unlike GetDisplayList which is required to return a value in [0.,1.]
  /// for each element of the display_list, this routine has the option of
  /// returning -1.  The caller can convert -1. to 0.5 if consistency with
  /// GetDisplayList is required.


  virtual INT4m GetDisplayList(REAL4m &xstep_request,
			       REAL4m &ystep_request,
			       REAL4m &zstep_request,
			       const Nb_BoundingBox<REAL4> &range,
			       const char * colorquantity,
			       Nb_List<Vf_DisplayVector> &display_list) =0;
  ///   Use ?step_request=0 to remove constraint on that axis, e.g., if
  /// xstep_request=ystep_request=zstep_request=0, then this routine places
  /// all points inside "range" into display_list.
  ///   Return value is number of points in display_list.  This is conceptually
  /// a const function, but isn't if the underlying representation stores the
  /// mesh values internally in a Nb_List<> structure, and uses the Nb_List<>
  /// iterator to access the elements (the iterator has an internal pointer
  /// that gets incremented with each access).  One might want to add a
  /// Nb_List<> "header" copy routine to the Nb_List<> template definition, to
  /// allow for efficient local iterators.  Then this function could be made
  /// const.
  ///   Note: All returned elements of display_list have a shade value in
  /// [0.,1.].  If colorquantity is unknown or can't be computed for some
  /// reason, then shade will be set to 0.5.

  virtual INT4m FindPreciseClosest(const Nb_Vec3<REAL8> &pos,
				   Nb_LocatedVector<REAL8> &lv) =0;
  // Finds closest vector in mesh to location loc;  The return value
  // is non-zero on error, which probably means the current mesh is
  // empty.  NOTE: This routine is likely not optimized for repeated
  // access.  For tight, time-critical loops, try to access the underlying
  // mesh using a mesh-specific lower level method.

  INT4m FindClosest(const Nb_Vec3<REAL4> &pos,
		    Nb_LocatedVector<REAL4> &lv);
  // Low precision version of the previous.

  Vf_Mesh(const char *_MeshTypename,const char *_Filename,
	  const char *_Title=NULL,const char* _Description=NULL,
	  const char *_MeshUnit=NULL,const char *_ValueUnit=NULL,
	  REAL8 _ValueMultiplier=1.):
    MeshTypename(_MeshTypename),Filename(NULL),
    Title(NULL),Description(NULL),
    MeshUnit(NULL),ValueUnit(NULL),
    UseMagHints(0),MinMagHint(0.),MaxMagHint(0.),
    ValueMultiplier(_ValueMultiplier) {
      SetFilename(_Filename);
      SetTitle(_Title);
      SetDescription(_Description);
      SetMeshUnit(_MeshUnit);
      SetValueUnit(_ValueUnit);
  }
  virtual INT4m GetSize() const =0;

  virtual ~Vf_Mesh() {
    if(Filename!=NULL)    delete[] Filename;
    if(Title!=NULL)       delete[] Title;
    if(Description!=NULL) delete[] Description;
    if(MeshUnit!=NULL)    delete[] MeshUnit;
    if(ValueUnit!=NULL)   delete[] ValueUnit;
  }

  // Use these to iterate through entire mesh.  The GetFirstPt() routine
  // constructs an appropriate Vf_Mesh_Index pointer, and returns the
  // address in 'key'.  Subsequent calls use this key to get the next
  // element from the mesh.  The return value of GetFirstPt and GetNextPt
  // is 0 if there are no more elements, 1 othersize.  The calling routine
  // is responsible for calling the destructor on 'key' when finished.
  // NOTE: Unlike some other generic Vf_Mesh functions, these return
  //       REAL8 vectors.
  virtual int
  GetFirstPt(Vf_Mesh_Index* &key,Nb_LocatedVector<REAL8> &) const {
    key=new Vf_Mesh_Index;
    return 1;
  }
  virtual int
  GetNextPt(Vf_Mesh_Index* &,Nb_LocatedVector<REAL8> &) const {
    return 1;
  }

};

//////////////////////////////////////////////////////////////////////////
// Empty mesh
class Vf_EmptyMesh:public Vf_Mesh
{
private:
  Nb_BoundingBox<REAL8> empty_range;
  Nb_List< Nb_Vec3<REAL4> > empty_boundary_list;
public:
  Vf_EmptyMesh():Vf_Mesh("Vf_EmptyMesh","<empty>") {}
  INT4m GetPreciseRange(Nb_BoundingBox<REAL8> &range) const {
    range=empty_range; return 0;
  }
  INT4m GetPreciseDataRange(Nb_BoundingBox<REAL8> &range) const {
    return GetPreciseRange(range);
  }
  INT4m GetBoundaryList(Nb_List< Nb_Vec3<REAL4> > &boundary_list) const {
    boundary_list=empty_boundary_list;
    return 0;
  }
  REAL4m GetApproximateCellWidth() const { return REAL4m(1.); }
  Nb_Vec3<REAL4>  GetApproximateCellDimensions() const {
    return Nb_Vec3<REAL4>(REAL4(1.),REAL4(1.),REAL4(1.));
  }
  REAL4m GetPreferredStepSize(REAL4m step) const {
    return REAL4m(OC_ROUND(step));
  }
  REAL4m GetSubsampleGrit(void) const { return REAL4m(1.); }
  INT4m GetDisplayList(REAL4m &,REAL4m &,REAL4m &,
		       const Nb_BoundingBox<REAL4>&,
		       const char *,
		       Nb_List<Vf_DisplayVector> &display_list) {
    display_list.Clear();
    return 0;
  }
  virtual INT4m FindPreciseClosest(const Nb_Vec3<REAL8> &,
				   Nb_LocatedVector<REAL8> &) {
    return 1;  // Not supported.
  }
  virtual INT4m GetSize() const { return 0; }
};

//////////////////////////////////////////////////////////////////////////
// Regular 3D grid of Nb_Vec3<REAL8>'s
class Vf_GridVec3f_Index:public Vf_Mesh_Index
{
public:
  int i,j,k;
  Vf_GridVec3f_Index() : i(0),j(0),k(0) {}
};

class Vf_GridVec3f:public Vf_Mesh
{ // NOTE: dim1,dim2,dim3 of grid should be thought of in the order z,y,x
  // (or k,j,i), because in applications typically z-dim (k) will be smallest,
  // x-dim (i) will be largest, so the data is stored arr[k][j][i].  This is
  // why we overload the () operator to map (i,j,k) -> [k][j][i].
  // IN PARTICULAR, efficient loops will increment the innermost loop on i.
  //   SO, KEEP THIS STRAIGHT: (i,j,k) coming in to member functions
  // should access grid[k][j][i], or (better) use GridVec(i,j,k).
private:
  static const ClassDoc class_doc;
  Nb_Array3D< Nb_Vec3<REAL8> > grid;   // Magnetization vector array
  inline INT4m SetSize(INT4m newi,INT4m newj,INT4m newk) {
    return grid.Allocate(newk,newj,newi);
  }
  Nb_BoundingBox<REAL8> range;
  Nb_BoundingBox<REAL8> data_range;
  Nb_List< Nb_Vec3<REAL8> > boundary;
  Nb_Vec3<REAL8> coords_base,coords_step;  
  Nb_Vec3<REAL8> NodalToProblemCoords(Nb_Vec3<REAL8> const &w_nodal) const;
  Nb_Vec3<REAL8> ProblemToNodalCoords(Nb_Vec3<REAL8> const &v_problem) const;
  /// The conversion from nodes to problem coordinates is
  ///    (i,j,k) -> base+(i*step.x,j*step.y,k*step.z).
  /// NOTE 1: The grid vector access functions, e.g., operator(), access
  ///   via the integral "nodal" coordinates, *not* in terms of the
  ///   (real valued) problem coordinates.
  /// NOTE 2: The values in coords_step are guaranteed to be !=0,
  ///   but may be negative!  However, GetStdStep() (below), returns
  ///   a value >0.
  /// NOTE 3: The coords conversion routines don't preserve order
  ///   if step.? is negative, e.g., say (min.x,min.y) < (max.x,max.y)
  ///   going in, but step.x is negative.  Then after conversion,
  ///   min.y~ will still be < max.y~, but now min.x~>max.x~.

  REAL8m GetStdStep() const {
    REAL8m basestep=OC_MIN(REAL8m(fabs(coords_step.x)),
			   REAL8m(fabs(coords_step.y)));
    return basestep;
  }

  // The following function is left undefined on purpose
  Vf_GridVec3f& operator=(Vf_GridVec3f& rhs); // Assignment op

public:
  INT4m GetDimens(INT4m &isize,INT4m &jsize,INT4m &ksize) const {
    return grid.GetSize(ksize,jsize,isize);
  }
  virtual INT4m GetSize() const { return grid.GetSize(); }
  Vf_GridVec3f():Vf_Mesh("Vf_GridVec3f",NULL,NULL) {}

  // The following constructor should be called from file input functions.
  Vf_GridVec3f(const char *filename,const char *title,const char *desc,
	       const char *meshunit,
	       const char *valueunit,REAL8 valuemultiplier,
               INT4m newi,INT4m newj,INT4m newk,
               const Nb_Vec3<REAL8> &basept,const Nb_Vec3<REAL8> &gridstep,
               const Nb_List< Nb_Vec3<REAL8> > &newboundary,
               const Nb_BoundingBox<REAL8> &newdatarange,
               const Nb_BoundingBox<REAL8> &newrange);

  Nb_Vec3<REAL8> GetBasePoint() const { return coords_base; }
  Nb_Vec3<REAL8> GetGridStep() const { return coords_step; }

  Vf_GridVec3f(Vf_GridVec3f &other):Vf_Mesh(NULL,"Vf_GridVec3f") {
    // Copy constructor
    SetFilename(other.GetName());
    SetTitle(other.GetTitle());
    SetDescription(other.GetDescription());
    SetMeshUnit(other.GetMeshUnit());
    SetValueUnit(other.GetValueUnit());
    ValueMultiplier=other.ValueMultiplier;
    grid.Copy(other.grid);
    data_range=other.data_range;
    range=other.range;
    boundary=other.boundary;
    coords_base=other.coords_base;
    coords_step=other.coords_step;
  }

  inline Nb_Vec3<REAL8>& GridVec(INT4m i,INT4m j,INT4m k) {
    return grid(k,j,i);
  }
  inline Nb_Vec3<REAL8>& operator()(INT4m i,INT4m j,INT4m k) {
    return GridVec(i,j,k);
  }
  inline const Nb_Vec3<REAL8>& GridVec(INT4m i,INT4m j,INT4m k) const {
    return grid(k,j,i);
  }
  inline const Nb_Vec3<REAL8>& operator()(INT4m i,INT4m j,INT4m k) const {
    return GridVec(i,j,k);
  }
  INT4m GetPreciseRange(Nb_BoundingBox<REAL8> &myrange) const;
  INT4m GetPreciseDataRange(Nb_BoundingBox<REAL8> &myrange) const;
  INT4m GetPreciseBoundaryList(Nb_List< Nb_Vec3<REAL8> >
			       &boundary_list) const {
    boundary_list=boundary;
    return 0;
  }

  virtual
  INT4m GetBoundaryList(Nb_List< Nb_Vec3<REAL4> > &boundary_list) const;

  virtual REAL4m GetApproximateCellWidth() const {
    return sqrt((OC_SQ(coords_step.x)+OC_SQ(coords_step.y))/2.);
    /// NOTE: This use to just return GetStdStep(), but the above
    /// is consistent with what is returned by GeneralMesh3f.
  }
  Nb_Vec3<REAL4>  GetApproximateCellDimensions() const {
    return Nb_Vec3<REAL4>(REAL4(fabs(coords_step.x)),
			  REAL4(fabs(coords_step.y)),
			  REAL4(fabs(coords_step.z)));
  }
  REAL4m GetPreferredStepSize(REAL4m step) const {
    REAL4m basestep=GetStdStep();
    return basestep*REAL4m(OC_ROUND(step/basestep));
  }
  REAL4m GetSubsampleGrit(void) const {
    return GetStdStep()/GetApproximateCellWidth();
  }
  virtual INT4m ColorQuantityTypes(Nb_List<Nb_DString> &types) const;
  virtual REAL4m GetVecShade(const char* colorquantity,
                             const Nb_Vec3<REAL4>& v) const;
  virtual INT4m GetDisplayList(REAL4m &xstep_request,
			       REAL4m &ystep_request,
			       REAL4m &zstep_request,
			       const Nb_BoundingBox<REAL4> &dsprange,
			       const char *colorquantity,
			       Nb_List<Vf_DisplayVector> &display_list);
  virtual INT4m FindPreciseClosest(const Nb_Vec3<REAL8> &pos,
				   Nb_LocatedVector<REAL8> &lv);
  virtual ~Vf_GridVec3f(void) {}

  virtual int
  GetFirstPt(Vf_Mesh_Index* &key,Nb_LocatedVector<REAL8> &lv) const;

  virtual int
  GetNextPt(Vf_Mesh_Index* &key,Nb_LocatedVector<REAL8> &lv) const;

};

//////////////////////////////////////////////////////////////////////////
// General (non-regular) 3D mesh of Nb_Vec3<REAL8>'s
class Vf_GeneralMesh3f_Index:public Vf_Mesh_Index
{
public:
  Vf_BoxList_Index key;
};

class Vf_GeneralMesh3f:public Vf_Mesh
{ 
private:
  static const ClassDoc class_doc;
  Vf_BoxList mesh;
  Nb_BoundingBox<REAL8> data_range;  // Minimal box containing data points
  Nb_BoundingBox<REAL8> range;       // Includes data_range+boundary
  Nb_List< Nb_Vec3<REAL8> > boundary;    // 2D boundary vertices
  Nb_Vec3<REAL8> step_hints;        // Hints on grid structure

  // The following functions are left undefined on purpose
  Vf_GeneralMesh3f& operator=(Vf_GeneralMesh3f& rhs); // Assignment op
  Vf_GeneralMesh3f(Vf_GeneralMesh3f&);  // Copy constructor

  // Provided for convience for mesh arrays.
  Vf_GeneralMesh3f(void):Vf_Mesh("Vf_GeneralMesh3f",NULL) {}
public:
  ////////////// Generic Mesh functions ///////////////////
  Vf_GeneralMesh3f(const char* filename,const char* title=NULL,
                   const char* desc=NULL,
		   const char* meshunit=NULL,
		   const char* valueunit=NULL,
		   REAL8 valuemultiplier=1.0)
    :Vf_Mesh("Vf_GeneralMesh3f",filename,title,desc,
	     meshunit,valueunit,valuemultiplier) {}
  /// The preceding constructor should be called from file input functions.

  void SetMeshTitle(const char* title) { SetTitle(title); }
  void SetMeshDescription(const char* desc) { SetDescription(desc); }

  virtual ~Vf_GeneralMesh3f(void) {}

  INT4m GetSize() const { return mesh.GetSize();  }
  INT4m GetPreciseRange(Nb_BoundingBox<REAL8> &myrange) const;
  INT4m GetPreciseDataRange(Nb_BoundingBox<REAL8> &myrange) const;
  INT4m GetPreciseBoundaryList(Nb_List< Nb_Vec3<REAL8> >
			       &boundary_list) const {
    boundary_list=boundary;
    return 0;
  }
  virtual
  INT4m GetBoundaryList(Nb_List< Nb_Vec3<REAL4> > &boundary_list) const;
  virtual REAL4m GetApproximateCellWidth() const;
  Nb_Vec3<REAL4>  GetApproximateCellDimensions() const;
  REAL4m GetPreferredStepSize(REAL4m step) const {
    REAL4m cellsize=GetApproximateCellWidth();
    return cellsize*REAL4m((OC_ROUND(step/cellsize)));
  }
  REAL4m GetSubsampleGrit(void) const { return REAL4m(0.1); }
  /// This is suppose to be a sensible lower bound on the sampling
  /// period relative to "GetApproximateCellWidth()".  This should
  /// probably be set to the smallest neighbor distance in the
  /// mesh, but for now we'll just use 0.1.
  virtual INT4m ColorQuantityTypes(Nb_List<Nb_DString> &types) const;
  virtual REAL4m GetVecShade(const char* colorquantity,
                             const Nb_Vec3<REAL4>& v) const;
  virtual INT4m GetDisplayList(REAL4m &xstep_request,
			       REAL4m & ystep_request,
			       REAL4m & zstep_request,
			       const Nb_BoundingBox<REAL4> &range_request,
			       const char *colorquantity,
			       Nb_List<Vf_DisplayVector> &display_list);

  virtual INT4m FindPreciseClosest(const Nb_Vec3<REAL8> &pos,
				   Nb_LocatedVector<REAL8> &lv);

  ////////////// Vf_GeneralMesh3f specific functions ///////////////////
  void ResetMesh(); // Resets data w/o changing filename, title or desc.
  void ResetMesh(const char* newFilename); // Resets data & filename
  void ResetMesh(const char* newFilename,const char* newTitle,
		 const char* newDescription);
  /// Resets data, filename, title and description.

  void AddPoint(const Nb_LocatedVector<REAL8>& lv);
  void SortPoints();   // Post point import processing, for efficient
  // accessing.  Will be called automatically by GetDisplayList() if needed,
  // but client access provided to allow this processing to be scheduled,
  // since this may take more than a few CPU cycles.
  void SetRange(const Nb_BoundingBox<REAL8> &newrange) { range=newrange; }
  void SetBoundaryList(const Nb_List< Nb_Vec3<REAL8> > &boundary_list);
  void SetApproximateStepHints();
  void SetStepHints(const Nb_Vec3<REAL8> &_step) { step_hints=_step; }
  Nb_Vec3<REAL8> GetStepHints() const { return step_hints; }
  virtual int
  GetFirstPt(Vf_Mesh_Index* &key,Nb_LocatedVector<REAL8> &lv) const;

  virtual int
  GetNextPt(Vf_Mesh_Index* &key,Nb_LocatedVector<REAL8> &lv) const;

};

#endif  // _VF_MESH
